# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# (c) 2007, 2008 Red Hat. 
# Written by skvidal@fedoraproject.org, wwoods@redhat.com

# TODO: gpgcheck downloaded pkgs
# TODO: probably have some defensive checks before we start anything
#       system sanity checks, depsolver checks against current rpmdb
# TODO: deal with conflicts and try to walk around them

import sys
import os
import os.path
import shutil
import ConfigParser
import shlex
import tempfile
from urlgrabber.grabber import URLGrabber, URLGrabError

import yum, rpm
import yum.Errors
import rpmUtils.arch
from yum.misc import checksum

from yum.parser import varReplace
from yum.constants import *

__version__ = "1.0.0"

# this should be configurable... 
if os.path.exists("releases.txt"):
    releasesfn = "releases.txt"
elif os.path.exists("data/releases.txt"):
    releasesfn = "data/releases.txt"
else:
    releasesfn = "http://mirrors.fedoraproject.org/releases.txt"

runfile = "/var/run/preupgrade"


def isParaVirt():
    """
    This function determines if the running system is a paravirt guest.
    This code is borrowed heavily from rhnreg.py (packaged with rhn-setup).
       2.  Check /sys/hypervisor/uuid.  If exists and is non-zero, we know
           the system is a para-virt guest; exit.
    """

    isGuest = False
    if os.path.exists("/proc/xen/xsd_port"):
        return False
    if os.path.exists("/sys/hypervisor/uuid"):
        try:
            uuid_file = open('/sys/hypervisor/uuid', 'r')
            uuid = uuid_file.read()
            uuid_file.close()
            if len(uuid.strip("0-\n")) != 0:
                isGuest = True
        except IOError:
            # Failed; must not be para-virt.
            pass

    return isGuest

def getlang():
    lang='en_US' # fallback
    infile = '/etc/sysconfig/i18n'
    s = shlex.shlex(open(infile),infile,posix=True)
    while True:
        t = s.get_token()
        if not t:
            break
        if t == 'LANG':
            eq = s.get_token() # =
            if eq == '=':
                lang = s.get_token()  # value
    return lang

def hrsize(size, precision=1):
    '''Return a string representing the "human-readable" size'''
    abbr = [(1<<40L,'TB'),
            (1<<30L,'GB'),
            (1<<20L,'MB'),
            (1<<10L,'kB'),
            (1,'bytes')]
    if size == 1:
        return "1 byte"
    for factor, suffix in abbr:
        if size > factor:
            break
    if precision == 0:
        return "%d%s" % (int(size)/factor,suffix)
    else:
        return ("%."+str(precision)+"f%s") % (float(size)/factor,suffix)

def st_dev_to_devicename(st_dev):
    '''Try to convert st_dev to a devicename by reading /proc/partitions.
    Obviously this will only work when st_dev represents a partition.'''
    (major,minor) = (os.major(st_dev),os.minor(st_dev))
    proc_partitions = open('/proc/partitions')
    dummy = proc_partitions.readline() # skip the header line
    majmin = [str(major),str(minor)]
    for line in proc_partitions:
        l = line.split()
        if len(l) == 4 and l[0:2] == majmin:
            return l[3]
    return None

def devicename_to_st_dev(device):
    '''Try to convert a devicename to an st_dev by using os.stat.'''
    s = os.stat(device)
    if hasattr(s,"st_rdev"):
        return s.st_rdev

def device_to_uuid(device):
    '''Try to convert a devicename to a filesystem UUID.'''
    if not device.startswith('/dev/'):
        device = '/dev/' + device
    if not os.path.exists(device):
        return None
    # FIXME: handle exceptions/bad return values here
    p = os.popen('/sbin/blkid -s UUID -o value ' + device)
    line = p.readline()
    p.close()
    return line.strip()

def get_mdraidinfo(devname):
    '''If devname is a mdraid device, return a dict with the following keys:
    {'level':"RAID level", 'members':[list of active, in-sync member devices]}. Example:
    {'level':'raid1','members':['sda1','sdb1]}
    Otherwise, returns an empty dict.'''
    results = {}
    md_sysdir="/sys/block/%s/md/" % devname

    # Get the RAID level; if this fails we don't have a proper md dir
    # (and thus this probably isn't a RAID device at all)
    try:
        f = open(md_sysdir+"level")
    except IOError:
        return results
    results['level'] = f.readline().strip()
    f.close()
    # Yep, it's RAID. Let's read in the member devices.
    # Info about how this works comes from md.txt in kernel-docs
    n = 0
    rd = md_sysdir+"rd%i" % n
    members = []
    while os.path.exists(rd):
        dev = os.path.basename(os.readlink(rd))
        dev = dev.replace("dev-","",1)
        members.append(dev)
        n += 1
        rd = md_sysdir+"rd%i" % n
    results['members'] = members
    return results

def bootpath_to_anacondapath(somepath,UUID=False):
    '''Convert a path under /boot to a path that anaconda will understand.
    Returns a string anaconda's loader can use to find that path, e.g.:
    '/boot/upgrade/stage2.img' -> 'hd:sda1:/upgrade/stage2.img'

    If UUID is set to True, the returned string will use UUID= to specify
    the target device, e.g.: 'hd:UUID=[uuid]:/upgrade/stage2.img'
    This is much more reliable, but only supported in Anaconda 11.5.1.15
    (Fedora 10 Alpha) or later.
    
    Raises PUError if something goes wrong.'''
    st_dev = os.stat(somepath).st_dev
    bootdev = st_dev_to_devicename(st_dev)
    mtab = open("/etc/mtab")
    mountpoint = ''
    for line in mtab:
        (dev,mp,rest) = line.split(' ',2)
        if os.path.exists(dev):
            if devicename_to_st_dev(dev) == st_dev: 
                mountpoint=mp
                break
    if not mountpoint:
        raise PUError, "Could not find mount point for boot device %s" % bootdev

    raid = get_mdraidinfo(bootdev)
    if raid:
        if raid['level'] == 'raid1':
            bootdev = raid['members'][0]
        else:
            raise PUError, "/boot is on unbootable RAID level '%s'" % raid['level']

    if mountpoint != '/':
        somepath = somepath.replace(mountpoint,'',1)

    if bootdev and somepath:
        if UUID:
            bootdev='UUID=%s' % device_to_uuid(bootdev)
        return 'hd:%s:%s' % (bootdev,somepath)
    else:
        return None

def get_sysconfig_keymap():
    '''Return the value of KEYTABLE in /etc/sysconfig/keyboard, or None
    if the file is missing or KEYTABLE is not set'''
    layout = None
    infile = '/etc/sysconfig/keyboard'
    try:
        s = shlex.shlex(open(infile),infile,posix=True)
        while s:
            t = s.get_token()
            if not t:
                break
            if t == 'KEYTABLE':
                eq = s.get_token() # =
                if eq == '=':
                    layout = s.get_token()  # value
    except IOError:
        pass
    return layout

def getkeymap():
    '''Get the system keymap from /etc/sysconfig/keyboard'''
    # Most upgrades won't require typing, so it's OK to just fall back to 'us'
    return get_sysconfig_keymap() or 'us'
    
def grubby(kernel,initrd,title="upgrade fedora",args='',):
    '''Run /sbin/grubby to create an entry for the given kernel+initrd.
    Overwrites any existing entries for the given kernel path.'''
    cmd='/sbin/grubby --title="%s" --remove-kernel="%s" --add-kernel="%s" --initrd="%s"' % (title,kernel,kernel,initrd)
    if args != '':
        cmd=cmd+' --args="%s"' % args
    #DEBUG FIXME
    print 'DEBUG %s' % cmd
 
    return os.system(cmd)
    
def grubby_make_next_boot_only():
    '''Tell grubby to make the newly-added boot entry the default for the
    next boot only.'''
    return os.system("/bin/echo 'savedefault --default=0 --once' | /sbin/grub >/dev/null")

def grubby_make_default(kernel):
    '''Make the given kernel path the default kernel for subsequent boots'''
    return os.system("/sbin/grubby --set-default=%s" % kernel)



class PUError(Exception):
    def __init__(self, value=None):
        Exception.__init__(self)
        self.value = value
    def __str__(self):
        return "%s" %(self.value,)

class PreUpgrade(yum.YumBase):
    
    def __init__(self, release=None, configfile=releasesfn):
        yum.YumBase.__init__(self)
        # FIXME what happens if the network isn't up yet?
        self.release_map = self.retrieve_release_info(configfile)
        self.bootpath='/boot/upgrade'
        self.image_csums = {} # filename = (csumtype, csum)
        # Noisy if these plugins don't exist, but doesn't raise an exception
        self._getConfig(enabled_plugins=['blacklist','whiteout'])
        if release:
            self.setup(release)

    def upgrade_available(self, want_stable=True):
        '''If there is a newer release available in the release map, return its
        name/key. If there are multiple newer releases, it returns the newest
        one (the one with the highest version number).
        Otherwise returns None.
        You can set stable to False if you want to consider non-stable releases
        as well (but right now this always gets you Rawhide).'''
        current = self.conf.yumvar['releasever']
        target_ver = current
        target_key = None
        for release_key,release in self.release_map.iteritems():
            # skip anything that's not marked preupgrade-ok
            if not self.release_preupgrade_ok(release):
                continue
            # skip unstable releases (unless the user has requested otherwise)
            if want_stable and not self.release_is_stable(release):
                continue
            # FIXME: unless we return a list or something, we'll always return 
            # Rawhide if stable=False, since it's always available and always
            # considered newer than whatever you've got.
            ver = release['version']
            if ver > target_ver: 
                target_ver = ver
                target_key = release_key
        return target_key

    def setup(self, release_key):
        if self.release_map.has_key(release_key):
            rel = self.release_map[release_key]
            self.pu_release = release_key

            self.pu_mirrorlist        = rel.get('mirrorlist')
            self.pu_baseurl           = rel.get('baseurl')
            self.pu_installmirrorlist = rel.get('installmirrorlist')
            self.pu_installurl        = rel.get('installurl')
            if not (self.pu_installmirrorlist or self.pu_installurl):
                self.pu_installmirrorlist = self.pu_mirrorlist
                self.pu_installurl = self.pu_baseurl
        else:
            raise PUError, 'Release does not match any in configuration.'

    def retrieve_release_info(self, configfile=None):
        releases = {}
        cp = ConfigParser.ConfigParser()
        if '://' in configfile:
            grabber = URLGrabber()
            configdata = grabber.urlopen(configfile)
        else:
            configdata = open(configfile)
        cp.readfp(configdata)
        for sect in cp.sections():
            rel = dict(cp.items(sect))
            if ('baseurl' in rel) or ('mirrorlist' in rel):
                releases[sect] = rel
                
        return releases
    
    def __release_key_is_true(self,release,key):
        '''Return True if the given key in the given release dict is a True
        value - something like 'true' or 'yes' or '1'.'''
        r = False
        val = release.get(key)
        if type(val) == str:
            b = val.lower()
            if b == '1' or b == 'true' or b == 'yes':
                r = True
        return r

    def release_is_stable(self,release):
        '''Return True if the given release is a stable release'''
        # this is an abstraction - just in case we change the structure of the
        # 'release' data.
        return self.__release_key_is_true(release,'stable')

    def release_preupgrade_ok(self,release):
        '''Return True if the given release is OK for preupgrade'''
        return self.__release_key_is_true(release,'preupgrade-ok')
        
    def setup_update_repo(self, download_progressbar=None):
        # disable all the other repos
        self.repos.disableRepo('*')
        # make our new repo obj
        newrepo = yum.yumRepo.YumRepository('preupgrade')
        newrepo.name = 'Anaconda Upgrade - %s' % self.pu_release
        if self.pu_mirrorlist: 
            self.pu_mirrorlist = varReplace(self.pu_mirrorlist, self.conf.yumvar)
        newrepo.mirrorlist = self.pu_mirrorlist
        newrepo.baseurl = []
        if self.pu_baseurl: 
            self.pu_baseurl = varReplace(self.pu_baseurl, self.conf.yumvar)
            newrepo.baseurl = self.pu_baseurl
        newrepo.enablegroups = 1
        
        # Set up a second yumRepo for boot images etc.)
        # (this is used by _retrieve_file)
        instrepo = yum.yumRepo.YumRepository('preupgrade-install-images')
        instrepo.name = 'Anaconda Upgrade (installable tree) - %s' % self.pu_release
        if self.pu_installmirrorlist:
            self.pu_installmirrorlist = varReplace(self.pu_installmirrorlist,self.conf.yumvar)
        instrepo.mirrorlist = self.pu_installmirrorlist
        instrepo.baseurl = []
        if self.pu_installurl:
            self.pu_installurl = varReplace(self.pu_installurl,self.conf.yumvar)
            instrepo.baseurl = self.pu_installurl

        newrepo.basecachedir = '/var/cache/yum/'
        instrepo.basecachedir = '/var/cache/yum/'

        # add our new repos
        self.repos.add(newrepo)        
        self.repos.add(instrepo)
        # enable the main repo
        self.repos.enableRepo(newrepo.id)
        # setup the progress bars for these repos
        if download_progressbar:
            newrepo.setCallback(download_progressbar)
            instrepo.setCallback(download_progressbar)
        # setup the repo dirs/etc
        self.doRepoSetup(thisrepo=newrepo.id)
        self.doRepoSetup(thisrepo=instrepo.id)
        self.distrepo = self.repos.findRepos(newrepo.id)[0]
        self.instrepo = self.repos.findRepos(instrepo.id)[0]
        # turn on caching
        self.conf.keepcache = 1
        # but force a check of the cached data
        self._cleanFiles(['cachecookie','mirrorlist.txt'],'cachedir','metadata')
        # This is a little dicey, using private stuff from yumRepo. Oh well.
        self.instgrab = self.instrepo._getgrab()
        # write self.pu_release to a file that resuming_run can pick up
        rf = open(runfile,"w")
        rf.write(self.pu_release + "\n")
        rf.close()

    def clear_update_repo(self):
        '''Clear out previously-set up repos (so we can retry)'''
        self.instrepo = None
        self.distrepo = None
        for id in ('preupgrade','preupgrade-install-images'):
            if self.repos.findRepos(id):
                self.repos.disableRepo(id)
                self.repos.delete(id)
    
    def retrieve_treeinfo(self):
        # take our repo
        # find a .treeinfo in them
        # download it
        # parse it
        # store relative path to the right kernel and initrd
        # there can be only one
        # info we need:
        # initrd.img path
        # kernel path
        # stage2.img path
        # min image
        # example:
        #   http://download.fedora.redhat.com/pub/fedora/linux/development/i386/os/.treeinfo
        tif = self.instrepo.grab.urlopen('/.treeinfo')
        #cracked out ppc issues, etc
        archlist = [self.conf.yumvar['basearch']]
        if isParaVirt():
            archlist = ['xen']
        if rpmUtils.arch.getCanonArch().startswith('ppc64'):
            archlist = [rpmUtils.arch.getCanonArch(), self.conf.yumvar['basearch']]
        
        cp = ConfigParser.ConfigParser()
        try:
            cp.readfp(tif)
        except ConfigParser.MissingSectionHeaderError:
            # Generally this means we failed to download the file
            raise PUError, "Failed to download installer metadata"
        except ConfigParser.Error:
            raise PUError, "Failed to parse installer metadata"
        
        mysection = None
        for arch in archlist:
            if cp.has_section('images-%s' % arch):
                mysection = 'images-%s' % arch
                break
        
        if not mysection:
            raise PUError, "Cannot find any kernel/initrd section for this arch in treeinfo"
            
        try:
            self.kernel = cp.get(mysection, 'kernel')
            self.initrd = cp.get(mysection, 'initrd')
            if cp.has_section('stage2'):
                self.mainimage = cp.get('stage2', 'mainimage')
            else:
                self.mainimage = 'images/stage2.img'
        except ConfigParser.NoOptionError, e:
            raise PUError, "Invalid treeinfo: %s" % str(e)
        
        if cp.has_section('checksums'):
            for opt_fn in cp.options('checksums'):
                (csum_type, csum) = cp.get('checksums', opt_fn).split(':')
                self.image_csums[opt_fn] = (csum_type, csum)
 
        tif.close()
        
    
    def _retrieve_file(self, fileinfo, targetdir=''):
        """ for returning files from our important repo, but not package or metadata files"""
        if not targetdir:
            targetdir = self.bootpath
        
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)

        # for each file:
        # check if we have the filesize and check it out
        # if it is okay, then continue
        # if it is not okay then unlink it 
        # check for space to download it
        
        # local name
        item_fname = os.path.basename(fileinfo)
        local = targetdir + '/' + item_fname
        # remote size
        tmp = self.instgrab.urlopen(fileinfo)
        mysize = tmp.hdr.dict['content-length']
        tmp.close()

        # Check to see if we've already got the right file
        if os.path.exists(local):
            local_stat = os.stat(local)
            if int(local_stat.st_size) == int(mysize):
                # It's the right size - check checksum
                if self.image_csums.has_key(fileinfo):
                    csumtype, csum = self.image_csums[fileinfo]
                    if csum == checksum(csumtype, local):
                        print "%s checksum OK" % local
                        return True
                    else:
                        print "%s checksum failed - removing" % local
                        os.unlink(local)
            else:
                os.unlink(local)

        # if we are this far we're going to try to download it
        # check for available disk space
        dirstat = os.statvfs(targetdir)
        if (dirstat.f_bavail * dirstat.f_bsize) <= long(mysize):
            raise PUError,("Not enough space in %s to download %s." % \
                    (targetdir, item_fname))

        try:
            self.instrepo._getFile(relative=fileinfo, local=local)
        except URLGrabError, e:
            os.unlink(local)
            raise PUError, "failure downloading %s: %s" % (item_fname, e)
            
        if self.image_csums.has_key(fileinfo):
            csumtype, csum = self.image_csums[fileinfo]
            if csum != checksum(csumtype, local):
                os.unlink(local)
                raise PUError, "checksum failure downloading %s" % item_fname

        return True

    def remove_boot_files(self):
        '''Remove any previously-downloaded boot files.'''
        def rmtree_handler(func,path,exc_info):
            (e_type, e_value, e_traceback) = exc_info
            raise PUError, "Failed to remove %s: %s" % (path,e_value)
        if os.path.exists(self.bootpath):
            shutil.rmtree(self.bootpath,ignore_errors=False,\
                          onerror=rmtree_handler)

    def remove_boot_entry(self):
        kernel = os.path.join(self.bootpath,
                            os.path.basename(getattr(self,'kernel','vmlinuz')))
        os.system('/sbin/grubby --remove-kernel="%s"' % kernel)

    def remove_repo_cache(self):
        '''Remove the repo caches, if they exist.
        Will throw PUError if the repos are not yet set up.'''
        # Since we inherit from YumBase we've got the cleanXXX methods:
        #   self.clean{Headers,Packages,Sqlite,Metadata}
        if not (hasattr(self,'distrepo') or hasattr(self,'instrepo')):
            # we need to set up the repos before we can clean them!
            raise PUError, "Can't clear repo caches until repos are set up"
        self.cleanHeaders()
        self.cleanPackages()
        self.cleanSqlite()
        self.cleanMetadata()
        
    def resuming_run(self):
        '''If there is cached data from a previously-interrupted run of
        preupgrade, return the release name (self.pu_release)'''
        releasename = None
        try:
            f = open(runfile)
            releasename = f.readline().strip()
        except IOError:
            pass
        return releasename

    def clear_incomplete_run(self):
        '''Clear all the saved data for an incomplete run.'''
        release = self.resuming_run()
        assert release != None
        self.setup(release)
        self.setup_update_repo()
        self.instrepo.enable() # cleanXXX only works on enabled repos
        self.remove_repo_cache()
        self.remove_boot_files()
        self.remove_boot_entry()
        os.unlink(runfile)
        self.clear_update_repo()

    def retrieve_critical_boot_files(self):
        # relative tmp FIXME XXX
        
        for item in (self.kernel, self.initrd):
            self._retrieve_file(item)
        
    def retrieve_updates_img(self,targetdir=None):
        '''Searches some pre-defined URL paths and attempts to retrieve an
        updates.img. Returns the local filename if a file was downloaded, 
        otherwise returns None. Will throw an exception on errors.'''
        # TODO: more.. official place for this
        baseurls = ['http://wwoods.fedorapeople.org/preupgrade/%s/updates.img']
        ver = self.release_map[self.pu_release]['version']
        arch = self.conf.yumvar['basearch']
        paths = [ver,ver+'/'+arch]
        targetfile = os.path.join(targetdir or self.bootpath,'updates.img')
        for url in [base % path for base in baseurls for path in paths]:
            try:
                u = self.instgrab.grabber.urlgrab(url,filename=targetfile)
            except URLGrabError, e:
                continue
            return targetfile

    def retrieve_non_critical_files(self,targetdir=''):
        self._retrieve_file(self.mainimage,targetdir)
        self.retrieve_updates_img(targetdir=targetdir)
        return os.path.basename(self.mainimage)

    def generate_kickstart(self,targetfile=None,extra_cmds=""):
        '''Generate a simple kickstart to automate the upgrade.
        Returns a tuple of (bootdev,path) like bootpath_to_anacondapath()'''
        lang = getlang()
        keymap = getkeymap()
        if not targetfile:
            targetfile = os.path.join(self.bootpath,'ks.cfg')
        kernel = os.path.join(self.bootpath,
                            os.path.basename(getattr(self,'kernel','vmlinuz')))
        f = open(targetfile,'w')
        ks = ['# ks.cfg generated by preupgrade',
              'lang %s' % lang,
              'keyboard %s' % keymap,
              'bootloader --upgrade',
              'upgrade',
              'reboot',
              '%s' % extra_cmds,
              '\n%post',
              'grubby --remove-kernel=%s' % kernel,
              'rm -rf %s /var/cache/yum/preupgrade*' % self.bootpath,
              '%end\n']
        f.write('\n'.join(ks))
        f.close()
        return bootpath_to_anacondapath(targetfile,UUID=True)
        
    def get_download_pkgs(self):
       return map(lambda txmbr: txmbr.po,
               self.tsInfo.getMembersWithState(output_states=TS_INSTALL_STATES))

    def get_download_size(self,pkgs):
        localsize = 0
        totalsize = 0
        for pkg in pkgs:
            try:
                size = int(pkg.size)
                totalsize += size
                try:
                    if pkg.verifyLocalPkg():
                        localsize += size
                except:
                    pass
            except e:
                print "Error calculating download size for %s: %s" % \
                            (str(e),str(pkg))
        return(totalsize - localsize)

    def get_upgrade_size(self):
        '''Returns the (approximate) amount of space needed for anaconda to
        complete the upgrade, in bytes.'''
        # Why, yes, this is a made-up number. And yet it's still more accurate
        # than the results of the test transaction.
        return 500 * 1024 * 1024

    # XXX currently unused. It takes forever, the callback doesn't work, and
    # the results don't match anaconda's transaction test so it tends to give
    # us false positives and then the upgrade fails.
    def transaction_test_diskspace(self,callback=None):
        '''Run a test transaction to ensure that we'll have enough disk
        to complete the upgrade.
        
        Returns a dict of mountpoint:spaceneeded pairs.
        An empty dict indicates success.
        '''
        testcb = yum.RPMTransaction(self,test=True)
        if callback:
            testcb.display = callback
        else:
            testcb.display = yum.rpmtrans.SimpleCliCallBack
        # save our dsCallback
        dscb = self.dsCallback
        self.dsCallback = None # ssshhhhhh
        # Turn off all the problem filters *but* diskspace
        self.ts.setProbFilter(~rpm.RPMPROB_FILTER_DISKSPACE)
        # XXX do we need setColor (see anaconda yuminstall.py) here?
        self.populateTs(keepold=0)
        # Actually test the transaction
        self.ts.addTsFlag(rpm.RPMTRANS_FLAG_TEST)
        problems = self.ts.run(testcb.callback, '')
        # clean up, and restore the depcheck callback
        del self.ts
        self.dsCallback = dscb
        spaceneeded = {}
        if problems:
            for (desc, (problem_type, mount, need)) in problems:
                if problem_type == rpm.RPMPROB_DISKSPACE:
                    spaceneeded[mount] = need
        return spaceneeded

    def generate_repodata(self,dir,comps=None,callback=None):
        '''Run createrepo on the given directory.'''
        try:
            from createrepo import MDError
            generate_repodata = self.generate_repodata_f9
        except ImportError:
            sys.path.append('/usr/share/createrepo')
            from dumpMetadata import MDError
            generate_repodata = self.generate_repodata_f8

        try:
            generate_repodata(dir, comps, callback)
        except (MDError, IOError), e:
            # FIXME: Handle this a little more cleanly
            print e
            return False
        return True

    def generate_repodata_f9(self, dir, comps=None, callback=None):
        '''Run createrepo (from Fedora 9 or later) on the given directory.'''
        from createrepo import MetaDataGenerator, MetaDataConfig
        conf = MetaDataConfig()
        conf.directory = dir
        conf.database = True
        if comps:
            conf.groupfile = comps
        mdgen = MetaDataGenerator(config_obj=conf, callback=callback)
        mdgen.doPkgMetadata()
        mdgen.doRepoMetadata()
        mdgen.doFinalMove()

    def generate_repodata_f8(self, dir, comps=None, callback=None):
        '''Run createrepo (from Fedora 8) on the given directory.'''
        from genpkgmetadata import MetaDataGenerator, parseArgs
        args = ['--database']
        if comps:
            args.extend(['--groupfile',comps])
        args.append(dir)
        cmds, dirs = parseArgs(args)
        tempdir = os.path.join(cmds['outputdir'], cmds['tempdir'])
        finaldir = os.path.join(cmds['outputdir'], cmds['finaldir'])
        for d in (tempdir, finaldir):
            if os.path.isdir(d):
                shutil.rmtree(d)
            elif os.path.exists(d):
                os.unlink(d)
        os.mkdir(tempdir)
        mdgen = MetaDataGenerator(cmds)
        mdgen.doPkgMetadata(dirs[0])
        mdgen.doRepoMetadata()
        os.rename(tempdir,finaldir)

    def return_available_versions(self, cur_ver, beta_ok=False):
        versions = []
        for name, releasedata in self.release_map.iteritems():
            # Skip releases that aren't newer than this one
            if cur_ver >= float(releasedata.get('version')):
                continue
            # Skip releases that aren't preupgrade-ok
            if not self.release_preupgrade_ok(releasedata):
                continue
            # If beta_ok get all releases, otherwise just the stable ones
            if beta_ok or self.release_is_stable(releasedata):
                versions.append(name)
        return versions
        
    def figure_out_what_we_need(self):
        # grab the comps file, we'll need it later for anaconda
        if self.comps.compscount == 0:
            raise PUError, "Cannot retrieve group lists from any repository"
        # also try to get the gzipped comps file if it's there so that
        # anaconda doesn't have to later.  this is kind of ugly (#439232)
        for r in self.repos.listGroupsEnabled():
            try:
                r.retrieveMD('group_gz')
            except:
                pass

        # update all
        self.update()
        # NOTE: anaconda sets installonly_limit to 1. Do we need that?
        # build a transaction
        self.buildTransaction()
        # print out what we would be doing

    def add_boot_target(self,extra_args=''):
        mykernel  = self.bootpath  + '/' + os.path.basename(self.kernel)
        myinitrd = self.bootpath + '/' + os.path.basename(self.initrd)
        arch = self.conf.yumvar['basearch']
        argstr = "preupgrade"
        if extra_args: 
            argstr += " " + extra_args
        title="Upgrade to %s" % self.pu_release
        # yaboot (ppc) doesn't support spaces in titles
        if arch.startswith('ppc'):
            title = 'preupgrade' 
        #FIXME - do something with return codes below here :)

        grubby(mykernel, myinitrd, title=title, args=argstr)
        if arch.startswith('ppc'):
            # yaboot doesn't like spaces in titles much
            # ppc machines need to run ybin to activate changes
            os.system('/sbin/ybin > /dev/null')

    def change_boot_target(self): 
        arch = self.conf.yumvar['basearch']
        if arch.startswith('ppc'):
            # ppc machines need to run ybin --bootonce
            os.system('/sbin/ybin --bootonce preupgrade > /dev/null')
        else:
            # everyone else uses grub --once
            grubby_make_next_boot_only()
            
    def _main(self):
        '''Basically an example of how a run is supposed to proceed.'''
        self.setup('Rawhide')
        self.setup_update_repo()        
        # In this example, we always clear old data rather than reusing it
        self.remove_repo_cache()
        self.remove_boot_files()
        self.figure_out_what_we_need()
        downloadpkgs = []
        downloadpkgs = map(lambda txmbr: txmbr.po,
                           self.tsInfo.getMembersWithState(output_states=TS_INSTALL_STATES))
        del(self.tsInfo)
        
       
        problems = self.downloadPkgs(downloadpkgs) 

        # find tree info data 
        self.retrieve_treeinfo()
        
        # grab kernel and initrd and stage2
        try:
            self.retrieve_critical_boot_files()
        except PUError, e:
            print 'Fatal: %s' % e
            sys.exit(1)

        try:            
            self.retrieve_non_critical_boot_files()        
        except PUError:
            pass
        
        # make it bootable
        self.add_boot_target()
        self.change_boot_target()


